/************************************************************************/
/*  A Part of ARM Cortex-A9 GNU GDB Stub Software.			*/
/*   				Copyright (C) 2013 Toshio MURAI		*/
/* This program is free software; you can redistribute it and/or modify	*/
/* it under the terms of the GNU General Public License as published by	*/
/* the Free Software Foundation; either version 3 of the License, or	*/
/* (at your option) any later version.					*/
/*									*/
/* This program is distributed in the hope that it will be useful,	*/
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*/
/* GNU General Public License for more details.  			*/
/************************************************************************/
	.text
	.global	_start

_start:
#FIQ mode
	MOV	R1, #0xD1	/* #(CPSR_MD_SVC|CPSR_IRQ|CPSR_FIQ)~THUMB	*/
	MSR	CPSR,  R1	/* Set FIQ mode		C:0-7 F:24-31/SX	*/
	LDR	SP,  =_exceptionStack	/* Supervisor Task Context stack 	*/
#IRQ mode
	MOV	R1, #0xD2	/* #(CPSR_MD_IRQ|CPSR_IRQ|CPSR_FIQ)		*/
	MSR	CPSR,  R1	/* Set IRQ mode					*/
	LDR	SP,  =_exceptionStack	/* Interrupt Task Context stack 	*/
#Abort mode
	MOV	R1, #0xD7	/* #(CPSR_MD_ABRT|CPSR_IRQ|CPSR_FIQ)~THUMB	*/
	MSR	CPSR,  R1	/* Set Abort mode	CF SX			*/
	LDR	SP,  =_exceptionStack	/* Abort Task Context stack 		*/
#Undefined mode
	MOV	R1, #0xDB	/* #(CPSR_MD_IRQ|CPSR_IRQ|CPSR_FIQ)		*/
	MSR	CPSR,  R1	/* Set Undefined Instruction  mode		*/
	LDR	SP,  =_exceptionStack	/* Interrupt Task Context stack 	*/
#Supervisor mode
	MOV	R1, #0xD3	/* #(CPSR_MD_SVR|CPSR_IRQ|CPSR_FIQ)~THUMB	*/
	MSR	CPSR,  R1	/* Set Supervisor mode	CF SX			*/
	LDR	SP,  =_supervisorStack	/* Supervisor Task Context stack 	*/

# C program routine (Initialize GDB Environment and return to undefined Instruction.)
	BL	initializeMemory
	BL	main
# Clear (Initialize ) Genreral Registers
	MOV	R0,#0
	MOV	R1,#1
	MOV	R2,#2
	MOV	R3,#3
	MOV	R4,#4
	MOV	R4,#5
	MOV	R6,#6
	MOV	R7,#7
	MOV	R8,#8
	MOV	R9,#9
	MOV	R10,#10
	MOV	R11,#11
	MOV	R12,#12
#	MOV	R13,#0		/* Do not Changem the Stack Pointer	*/
	MOV	R14,#.LOOP+4
#	MOV	R15,#0		/* Do not Changem the Program Counter	*/

.LOOP:	/*	Go to Undefine Instruction to handleException Funciotn	*/
	.word 0xe7ffdefe
	B	.LOOP

	.extern registers

.global	_undefinedInstruction
# Undefined Instruction :	Only Used GDB
_undefinedInstruction: 

	SUBS	LR,LR,#4
	STMDB	SP!,{R0-R12,R14}
	
	##########################
	LDR	R0 ,=registers
	MOV	R1 , SP
##
	STMIA	R0,{R0-R15}
# R0
	LDR	R2, [R1, #0]
	STR	R2, [R0, #0]
# R1
	LDR	R2, [R1, #4]
	STR	R2, [R0, #4]
# R15=PC
	STR	LR, [R0, #60]
# F0-F7 (12byte x 8) = 96 byte offset
# CPSR
	MRS	R1, SPSR	
	STR	R1, [R0, #164]
	###########################

# Interrupted mode (SVR/IRQ)	/* Only SVR ? which one is better?		*/
	MSR	CPSR, R1	/* If User Mode we cannot retrun		*/
# Supervisor mode R13(SP)/R14(LR)
#	MOV	R1, #0xD3	/* #(CPSR_MD_SVR|CPSR_IRQ|CPSR_FIQ)~THUMB	*/
#	MSR	CPSR,  R1	/* Set Supervisor mode	CF SX			*/

# R13 (SP)
	STR	sp, [r0, #52]
# R14 (LR)
	STR	r14, [r0, #56]
	
# undefined Instruction mode
	MOV	R1, #0xDB	/* #(CPSR_MD_IRQ|CPSR_IRQ|CPSR_FIQ)		*/
	MSR	CPSR,  R1	/* Set Undefined Instruction  mode		*/

	MOV     R0, #1
	BL      handleException

	LDR	R0 ,=registers
	MOV	R1 , sp
	LDR	R2, [R0, #60]
	STR	R2, [R1, #52]	//	PC-> R14/LR

#	LDR	R2, [R0, #0]
#	STR	R2, [R1, #0]	//	R0->R0
#	LDR	R2, [R0, #4]
#	STR	R2, [R1, #4]	//	R1->R1
#	LDR	R2, [R0, #8]
#	STR	R2, [R1, #8]	//	R2->R2
#	LDR	R2, [R0, #12]
#	STR	R2, [R1, #12]	//	R3->R3
#	LDR	R2, [R0, #16]
#	STR	R2, [R1, #16]	//	R4->R4
#	LDR	R2, [R0, #20]
#	STR	R2, [R1, #20]	//	R5->R5
#	LDR	R2, [R0, #24]
#	STR	R2, [R1, #24]	//	R6->R6
#	LDR	R2, [R0, #28]
#	STR	R2, [R1, #28]	//	R7->R7
#	LDR	R2, [R0, #32]
#	STR	R2, [R1, #32]	//	R8->R8
#	LDR	R2, [R0, #36]
#	STR	R2, [R1, #36]	//	R9->R9
#	LDR	R2, [R0, #40]
#	STR	R2, [R1, #40]	//	R10->R10
#	LDR	R2, [R0, #44]
#	STR	R2, [R1, #44]	//	R11->R11
#	LDR	R2, [R0, #48]
#	STR	R2, [R1, #48]	//	R12->R12
####	LDR	R2, [R0, #52]
####	STR	R2, [R1, #52]	//	R13->R13
####	LDR	R2, [R0, #52]
####	STR	R2, [R1, #56]	//	R14->R15

	LDMIA	SP!,{R0-R12, PC}^
#	LDMIA	SP!,{R0-R12,R14}
#	BX	LR
####	SUBS	PC, R14, #4

.global	_softwareInterrupt
_softwareInterrupt:
#	SUBS	LR,LR,#0
	STMDB	SP!,{R0-R12, R14}

#############################################
	LDR	R0 ,=registers
	MOV	R1 , SP
##
	STMIA	R0,{R0-R15}
# R0
	LDR	R2, [R1, #0]
	STR	R2, [R0, #0]
# R1
	LDR	R2, [R1, #4]
	STR	R2, [R0, #4]
# R15=PC
	STR	LR, [R0, #60]
# F0-F7 (12byte x 8) = 96 byte offset
# CPSR
	MRS	R1, SPSR	
	STR	R1, [R0, #164]
###############################################

	MOV     R0, #2
	BL      handleException
#	BL	softwareInterrupt
	LDMIA	SP!,{R0-R12, R14}
#	LDMIA	SP!,{R0-R12, PC}^

	STMDB	SP!, {R0,R1}
	LDR	R0,.vector
	LDR	R0,[R0]
#	mov	R0,#userVector
	ADD	R0, R0, #0x08
	STR	R0,[SP,#4]
	LDMIA	SP!, {R0,PC}

.global	_prefetchAbort
_prefetchAbort:	
	SUBS	LR,LR,#4
	STMDB	SP!,{R0-R12,R14}
	
	##########################
	LDR	R0 ,=registers
	MOV	R1 , SP
##
	STMIA	R0,{R0-R15}
# R0
	LDR	R2, [R1, #0]
	STR	R2, [R0, #0]
# R1
	LDR	R2, [R1, #4]
	STR	R2, [R0, #4]
# R15=PC
	STR	LR, [R0, #60]
# F0-F7 (12byte x 8) = 96 byte offset
# CPSR
	MRS	R1, SPSR	
	STR	R1, [R0, #164]
	###########################

# Interrupted mode (SVR/IRQ)	/* Only SVR ? which one is better?		*/
	MSR	CPSR, R1	/* If User Mode we cannot retrun		*/
# Supervisor mode R13(SP)/R14(LR)
#	MOV	R1, #0xD3	/* #(CPSR_MD_SVR|CPSR_IRQ|CPSR_FIQ)~THUMB	*/
#	MSR	CPSR,  R1	/* Set Supervisor mode	CF SX			*/

# R13 (SP)
	STR	sp, [r0, #52]
# R14 (LR)
	STR	r14, [r0, #56]
	
###############################################
# Abort mode
	MOV	R1, #0xD7	/* #(CPSR_MD_ABRT|CPSR_IRQ|CPSR_FIQ)~THUMB	*/
	MSR	CPSR,  R1	/* Set Abort mode	CF SX			*/

	MOV     R0, #3
	BL      handleException
#	BL	prefetchAbort

	LDR	R0 ,=registers
	MOV	R1 , sp
	LDR	R2, [R0, #60]
	STR	R2, [R1, #52]	//	PC-> R14/LR

	LDMIA	SP!,{R0-R12, PC}^
#	LDMIA	SP!,{R0-R12, R14}
#	STMDB	SP!, {R0,R1}
#	LDR	R0,.vector
#	LDR	R0,[R0]
#	ADD	R0, R0, #0x0C
#	STR	R0,[SP,#4]
#	LDMIA	SP!, {R0,PC}


.global	_dataAbort
_dataAbort:
	SUBS	LR,LR,#8
	STMDB	SP!,{R0-R12,R14}
	
	##########################
	LDR	R0 ,=registers
	MOV	R1 , SP
##
	STMIA	R0,{R0-R15}
# R0
	LDR	R2, [R1, #0]
	STR	R2, [R0, #0]
# R1
	LDR	R2, [R1, #4]
	STR	R2, [R0, #4]
# R15=PC
	STR	LR, [R0, #60]
# F0-F7 (12byte x 8) = 96 byte offset
# CPSR
	MRS	R1, SPSR	
	STR	R1, [R0, #164]
	###########################

# Interrupted mode (SVR/IRQ)	/* Only SVR ? which one is better?		*/
	MSR	CPSR, R1	/* If User Mode we cannot retrun		*/
# Supervisor mode R13(SP)/R14(LR)
#	MOV	R1, #0xD3	/* #(CPSR_MD_SVR|CPSR_IRQ|CPSR_FIQ)~THUMB	*/
#	MSR	CPSR,  R1	/* Set Supervisor mode	CF SX			*/

# R13 (SP)
	STR	sp, [r0, #52]
# R14 (LR)
	STR	r14, [r0, #56]
	
###############################################
# Abort mode
	MOV	R1, #0xD7	/* #(CPSR_MD_ABRT|CPSR_IRQ|CPSR_FIQ)~THUMB	*/
	MSR	CPSR,  R1	/* Set Abort mode	CF SX			*/

	MOV     R0, #4
	BL      handleException
#	BL	dataAbort

	LDR	R0 ,=registers
	MOV	R1 , sp
	LDR	R2, [R0, #60]
	STR	R2, [R1, #52]	//	PC-> R14/LR

	LDMIA	SP!,{R0-R12, PC}^	// will not work!
#	LDMIA	SP!,{R0-R12, R14}
#	STMDB	SP!, {R0,R1}
#	LDR	R0,.vector
#	LDR	R0,[R0]
#	ADD	R0, R0, #0x0C
#	STR	R0,[SP,#4]
#	LDMIA	SP!, {R0,PC}

.global	_interrupt
_interrupt:	
#	SUB	LR,LR,#4
	STMDB	SP!,{R0-R12, R14}
#############################################
	LDR	R0 ,=registers
	MOV	R1 , SP
##
	STMIA	R0,{R0-R15}
# R0
	LDR	R2, [R1, #0]
	STR	R2, [R0, #0]
# R1
	LDR	R2, [R1, #4]
	STR	R2, [R0, #4]
# R15=PC
	STR	LR, [R0, #60]
# F0-F7 (12byte x 8) = 96 byte offset
# CPSR
	MRS	R1, SPSR	
	STR	R1, [R0, #164]
###############################################
#	MOV     R0, #0x6	/* Arg#1 6 */
	BL      interrupt
	LDMIA	SP!,{R0-R12, R14}
#	LDMIA	SP!,{R0-R12, PC}^

	STMDB	SP!, {R0,R1}
	LDR	R0,.vector
	LDR	R0,[R0]
	ADD	R0, R0, #0x18
	STR	R0,[SP,#4]
	LDMIA	SP!, {R0,PC}


.global	_firstInterrupt
_firstInterrupt:	
#	SUB	LR,LR,#4
	STMDB	SP!,{R0-R12, R14}
#############################################
	LDR	R0 ,=registers
	MOV	R1 , SP
##
	STMIA	R0,{R0-R15}
# R0
	LDR	R2, [R1, #0]
	STR	R2, [R0, #0]
# R1
	LDR	R2, [R1, #4]
	STR	R2, [R0, #4]
# R15=PC
	STR	LR, [R0, #60]
# F0-F7 (12byte x 8) = 96 byte offset
# CPSR
	MRS	R1, SPSR	
	STR	R1, [R0, #164]
###############################################
#	MOV     R0, #0x7	/* Arg#1 6 */
#	bl      fiqException
	LDMIA	sp!,{r0-r12, r14}
#	LDMIA	sp!,{r0-r12, pc}^

	STMDB	SP!, {R0,R1}
	LDR	R0,.vector
	LDR	R0,[R0]
	ADD	R0, R0, #0x1C
	STR	R0,[SP,#4]
	LDMIA	SP!, {R0,PC}

#
#	Interrupt disable
#
.global	disableInterrupt
disableInterrupt:	/* IRQ Disable	FIQ 0x40*/
	MRS	R0, CPSR_all
	ORR	R0, R0 ,#0x080	/*#CPSR_IRQ*/
	MSR	CPSR_all, R0
	MOV	PC, LR		 /*  return */
#
#	Interrupt enable
#
.global	enableInterrupt
enableInterrupt:	/* IEQ Enable FIQ ~0x40*/
	MRS	R0, CPSR_all
	AND	R0, R0 ,#~0x80	/*#~CPSR_IRQ*/
	MSR	CPSR_all, R0
	MOV	PC, LR		 /*  return */

#
#	Coprocessor Support Routine
#

#MRC p15, 0, <Rd>, c12, c0, 0 ; Read VBAR Register
#MCR p15, 0, <Rd>, c12, c0, 0 ; Write VBAR Register
.global	readVectorBaseRegister
readVectorBaseRegister:
	MRC	P15, 0, R0, C12, C0, 0 
	MOV	PC, LR		 /*  return */
	
.global	writeVectorBaseRegister
writeVectorBaseRegister:
#	LDR	R0,#0x20200000
	MCR	P15, 0, R0, C12, C0, 0 
#	MRC	P15, 0, R0, C12, C0, 0 
	MOV	PC, LR		 /*  return */

#	High/Low Vector / Vbit 13th bit of SCTLR
#MRC p15, 0,<Rd>, c1, c0, 0; Read SCTLR
#MCR p15, 0,<Rd>, c1, c0, 0; Write SCTLR
.global	readSCTLR
readSCTLR:
	MRC	P15, 0,R0, C1, C0, 0
	MOV	PC, LR		 /*  return */

.global	setHighVector
setHighVector:
	MRC	P15, 0,R0, C1, C0, 0
	MOV	R1,#0x00002000
	ORR	R0,R1
	MCR	P15, 0,R0, C1, C0, 0
	MOV	PC, LR		 /*  return */
	
.global	setLowVector
setLowVector:
	MRC	P15, 0,R0, C1, C0, 0
	MOV	R1,#0xFFFFDFFF
	AND	R0,R1
	MCR	P15, 0,R0, C1, C0, 0
	MOV	PC, LR		 /*  return */

#
.vector:
	.word	vectorAddress

#vectorAddress:
#        .word   537919488
#        .word   0x20100000

##################################################################
#			/* N	  0b1000 0000 - - 0000 0000	*/
#			/* Z	  0b0100 0000 - - 0000 0000	*/
#			/* C	  0b0010 0000 - - 0000 0000	*/
#			/* V	  0b0001 0000 - - 0000 0000	*/
#			/* A	      0b1 0000 0000	*/
#			/* I		0b1000 0000	*/
#			/* F		0b0100 0000	*/
#			/* T		0b0010 0000	*/
#			/* USER		   0b1 0000	*/
#			/* FIQ		   0b1 0001	*/
#			/* IRQ		   0b1 0010	*/
#			/* SVR		   0b1 0011	*/
#			/* UDEF		   0b1 1011	*/
#			/* SYS		   0b1 1111	*/
