/************************************************************************/
/*  A Part of ARM Cortex-A9 GNU GDB Stub Software.			*/
/* Original version of this software comes from gdbstub for sparc in GDB*/ 
/* This program is free software; you can redistribute it and/or modify	*/
/* it under the terms of the GNU General Public License as published by	*/
/* the Free Software Foundation; either version 3 of the License, or	*/
/* (at your option) any later version.					*/
/*									*/
/* This program is distributed in the hope that it will be useful,	*/
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*/
/* GNU General Public License for more details.  			*/
/************************************************************************/

//***********************************************************************/
//*	GNU Linker .text/.bss Initialize routine before main function.	*/
//***********************************************************************/
extern char etext[], sdata[], edata[], sbss[] , ebss[];
void
initializeMemory()
{
	char *src, *dst;
	
	for( src = etext, dst = sdata ; dst< edata ; ) 
		*dst++ = *src++;
		
	for( dst=sbss                 ; dst< ebss  ; ) 
		*dst++ =0;
}

//***********************************************************************/
int
main ()
{

	initializeCPU();

	initializeUART();

	puts("\r\nWell come to RZ ARM CortexA9 GDB Stub.\r\n");

}

//***********************************************************************/
//*	Follwing Code Mainly comes from sparc veriosn of gdb stub.c	*/
//***********************************************************************/

static const char hexchars[] = "0123456789abcdef";

#define REGNUM	42
#define BUFMAX 512

int registers[REGNUM];		//R0=0 SP=13 PC=15 CPSR = 41
char remcomInBuffer[BUFMAX];
char remcomOutBuffer[BUFMAX];

char *
strcpy (char *dest, const char *src)
{
	while (*src != '\0') {
		*dest++ = *src++;
	}
	*dest = 0;
	return dest;
}

static int
hex (char ch)
{
	if ((ch >= 'a') && (ch <= 'f'))
		return (ch - 'a' + 10);
	if ((ch >= '0') && (ch <= '9'))
		return (ch - '0');
	if ((ch >= 'A') && (ch <= 'F'))
		return (ch - 'A' + 10);
	return (-1);
}

static char *
mem2hex (char *mem, char *buf, int count)
{
	int i;
	int ch;
	for (i = 0; i < count; i++) {
		ch = *mem++;
		*buf++ = hexchars[(ch>>4) & 0x0f];
		*buf++ = hexchars[ ch &0x0f];
	}
	*buf = 0;
	return (buf);
}

static char *
hex2mem (char *buf, char *mem, int count)
{
	int i;
	unsigned char ch;
	for (i = 0; i < count; i++) {
		ch = hex (*buf++) << 4;
		ch = ch + hex (*buf++);
		*mem++ = ch;
	}
	return (mem);
}

static int
hexToInt (char **ptr, int *intValue)
{
	int numChars = 0;
	int hexValue;
	*intValue = 0;
	while (**ptr) {
		hexValue = hex (**ptr);
		if (hexValue >= 0) {
			*intValue = (*intValue << 4) | hexValue;
			numChars++;
		}
		else
			break;
		(*ptr)++;
	}
	return (numChars);
}

char *
getpacket (void)
{
	unsigned char *buffer = &remcomInBuffer[0];
	unsigned char checksum;
	unsigned char xmitcsum;
	int count;
	char ch;

	while (1) {
		/* wait around for the start character, ignore all other characters */
		while ((ch = getDebugChar ()) != '$');

	      retry:
		checksum = 0;
		xmitcsum = -1;
		count = 0;

		/* now, read until a # or end of buffer is found */
		while (count < BUFMAX) {
			ch = getDebugChar ();
			if (ch == '$')
				goto retry;
			if (ch == '#')
				break;
			checksum = checksum + ch;
			buffer[count] = ch;
			count = count + 1;
		}
		buffer[count] = 0;
		if (ch == '#') {
			ch = getDebugChar ();
			xmitcsum = hex (ch) << 4;
			ch = getDebugChar ();
			xmitcsum += hex (ch);

			if (checksum != xmitcsum) {
				putDebugChar ('-');	/* failed checksum */
			}
			else {
				putDebugChar ('+');	/* successful transfer */

				/* if a sequence char is present, reply the sequence ID */
				if (buffer[2] == ':') {
					putDebugChar (buffer[0]);
					putDebugChar (buffer[1]);

					return &buffer[3];
				}

				return &buffer[0];
			}
		}
	}
}

/* send the packet in buffer. */
static void
putpacket (unsigned char *buffer)
{
	unsigned char checksum;
	int count;
	unsigned char ch;
	int check;

	/*  $<packet info>#<checksum>. */
	do {
		putDebugChar ('$');
		checksum = 0;
		count = 0;

		while (ch = buffer[count]) {
			putDebugChar (ch);
			checksum += ch;
			count += 1;
		}

		putDebugChar ('#');
		putDebugChar (hexchars[checksum >> 4]);
		putDebugChar (hexchars[checksum & 0xf]);

	} while (getDebugChar() != '+');
}

/*
##     0x00 	 Reset			0
##     0x04 	 Undefined Instruction	1
##     0x08 	 Software Interrupt	2
##     0x0c 	 Abort(prefetch)	3
##     0x10 	 Abort(data)		4
##     0x14 	 Reserved		5
##     0x18 	 IRQ			6
##     0x1c 	 FIQ			7
 */

#define	SIGHUP	1
#define	SIGINT	2
#define	SIGQUIT	3
#define	SIGILL	4
#define	SIGTRAP	5
#define	SIGABRT	6
#define	SIGENT	7
#define	SIGFPE	8
#define	SIGKILL	9
#define	SIGBUS	10
#define	SIGSEGV	11
#define	SIGSYS	12
#define	SIGPIPE	13
#define	SIGALRM	14
#define	SIGTERM	15
#define	SISSTOP	23

int
computeSignal (int exceptionVector)
{
	static int signal[] = {
		SIGINT,	/* Reset	*/	
		SIGILL,	/* Undefined Instruction	*/	
		SIGTRAP,	/* Software Interrupt	*/
		SIGSEGV,	/* Prefetch Abort	*/
		SIGBUS,		/* Data Abort	*/
		0,	/* Reserved	*/
		SIGINT,	/* IRQ		*/
		SIGINT,	/* FIQ		*/
	};

	return (signal[exceptionVector]);
}

//#define ILLEGAL_INST_ARM 0xe7ffdefe
void
handleException (int exceptionVector)
{
	int sigval;
	int addr, length;
	char *ptr;
	int regnum;
	unsigned int regval;

	/* reply to host that an exception has occurred */
	sigval = computeSignal (exceptionVector);
	if (sigval > 0) {
		remcomOutBuffer[0] = 'S';
		remcomOutBuffer[1] = hexchars[(sigval>>4) & 0x0f];
		remcomOutBuffer[2] = hexchars[ sigval &0x0f];
		remcomOutBuffer[3] = 0;

		putpacket (remcomOutBuffer);
	}

	clearStepBreakPoints();

	remcomOutBuffer[0] = 0;

	while (1) {
		remcomOutBuffer[0] = 0;
		ptr = getpacket ();

		switch (*ptr++) {
		case 'H':
		case '!':
			strcpy (remcomOutBuffer, "OK");
			break;
		case '?':
			remcomOutBuffer[0] = 'S';
			remcomOutBuffer[1] = hexchars[(sigval>>4) & 0x0f];
			remcomOutBuffer[2] = hexchars[ sigval &0x0f];
			remcomOutBuffer[3] = 0;
			break;
		case 'g':	/* return the value of the CPU registers */
			mem2hex ((char *) registers, remcomOutBuffer,
				 REGNUM*4);
			break;
		case 'G':	/* set the value of the CPU registers - return OK */
			hex2mem (ptr, (char *) registers, REGNUM*4);
			strcpy (remcomOutBuffer, "OK");
			break;

		case 'P':	/* set the value of the CPU registers with index- return OK */
			// 
			regnum = hex (*ptr);
			ptr++;
			ptr++;
			hex2mem (ptr, (char *) &regval, 4);
			registers[regnum] = regval;
			strcpy (remcomOutBuffer, "OK");
			break;

			/* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
		case 'm':
			/* TRY, TO READ %x,%x.  IF SUCCEED, SET PTR = 0 */
			if ( (hexToInt (&ptr, &addr))
					&& (*ptr++ == ',')
					&& (hexToInt (&ptr, &length)) ){
				if(mem2hex ((char *) addr, remcomOutBuffer, length)) break;
				strcpy (remcomOutBuffer, "E03");
			} else {
				strcpy (remcomOutBuffer, "E01");
			}
			break;
			/* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */
		case 'M':
			/* TRY, TO READ '%x,%x:'.  IF SUCCEED, SET PTR = 0 */
			if (hexToInt (&ptr, &addr)
					&& *ptr++ == ','
					&& hexToInt (&ptr, &length)
					&& *ptr++ == ':') {
				if(hex2mem (ptr, (char *) addr, length)){
					strcpy (remcomOutBuffer, "OK");
				}else {
					strcpy (remcomOutBuffer, "E03");
				}
			} else {
				strcpy (remcomOutBuffer, "E02");
			}
			break;

			/* sAA..AA   Step one instruction from AA..AA(optional) */
		case 'S':
		case 's':
			setStepBreakPoints();

			/* cAA..AA    Continue at address AA..AA(optional) */
		case 'C':
		case 'c':
			//enableUARTInterrupt ();
			return;
//		case 'Z': case 'z': case 'k': case 'R':	case 'r':
		default:
			break;
		}		/* switch */

		putpacket (remcomOutBuffer);

	}

	//enableUARTInterrupt();
	return;
}
