/************************************************************************/
/* CEV-RZ/A1l  GNU CC Control Software / (C) Toshio MURAI  2013		*/
/*      This is Free Software;   You can redistribute this software     */
/*      and/or modify it under the terms of the GNU General Public      */
/*      License version 3, or (at your option) any later version.       */
/************************************************************************/
#include 	"io.h"

void initializeCPU(){	
	*RMPR = 0x00;	// Address Remap

	setLowVector();		// setHighVector();

	if(readVectorBaseRegister()!=0x20200000) {
		writeVectorBaseRegister(0x20100000);
	}

}

initializeUART() {
	
	//	Initialize the P7_10,11 for SC3
	PMC->P7	|=	0x0C00	;
	PFCAE->P7	|=	0x0C00	;
	PFCE->P7	&= ~0x0C00	;
	PFC->P7	&= ~0x0C00	;	// 5th Fucntion mode ( TxD3/RxD3 )

	PM->P7		&= ~0x0800	;	// P7_11 TxD (0:out)
	PM->P7		|=  0x0400	;	// P7_10 RxD (1:in)

	{int i;for (i=100000; i>0 ; i--);} // wait for a while to send all charactors
	//	Initialize SC3 Async 115200 8NP Enable Interrupt BR/ER/RX/TX
	SC3->SCSCR = 00;	//	Control Register clear
	SC3->SCFCR = 06;	//	Reset FIFO 
	SC3->SCFCR = 00;	//

	SC3->SCBRR = 0x11;	//  0x11=115200  0x68=19200 baud register

	SC3->SCSMR = ((SC3->SCSMR & ~0xF8) | 0x00);		// Mode Register
	//	~0x80 Async	~0x40 8bit	~0x20 NoPa	~0x10 Even
	//	~0x08 1Stop	~0x04 Researve 	 0x03 1/64 0x02 1/16 0x01 1/4
	SC3->SCSCR = ((SC3->SCSCR & ~0x00) | 0xF8);		// Control Register
	//	 0x80 TIE	 0x40 RIE	 0x20 TE	 0x10 RE
	//	 0x08 REIE	~0x04 Researve	 0x00 Internal Clock 
	//i = SC3->SCFSR;	//
	SC3->SCFSR &= 0x00;	//
	//	 0xF000	Parity Error Numbrt	 0x0F00	Framing Error Numbrt
	//	 0x80 ERror	 0x40 TEND	 0x20 TDFEmpty	 0x10 BRK
	//	 0x0f FERror	 0x04 PERror	 0x02 RDFull	 0x01 (Recieve)DReasy

}

putch(volatile struct sc_str *p,int ch) {
	while(!(p->SCFSR & 0x20 ));	// TDFEmpty
	p->SCFTDR = ch;
	p->SCFSR &= ~0x60;	// Clear TEND/TDFE
}

getch(volatile struct sc_str *p){
	int	ch;
	if(p->SCFSR & 0x91) { p->SCFSR &= ~0x91; }//0x80 ER 0x10 BRK 0x01 DR
	if(p->SCLSR & 0xFF) { p->SCLSR = 0; }

	while(!(p->SCFSR & 0x02));	//  RDF
	ch = p->SCFRDR;
	p->SCFSR &= ~0x02;		// Clear RDF
	return ch;
}

putstr(char * str)
{
	while(*str) {
		if(*str =='\n')
			putch(SCPORT,'\r');
		putch(SCPORT,*str++);
	}
}

/*	On Board LED 	1 bit  P1.18 (L.on)	*/
initializeBoardLED(){
	PMC->P7	&= ~0x0300	;
	PM->P7	&= ~0x0300	;
	PORT->P7	|=  0x0100	;	// No.0 1:OFF  
	PORT->P7	&= ~0x0200	;	// No.1 0:ON (Center Side)
}

reverseLED(){PORT->P7^=0x0300;}
reverseLED0(){PORT->P7^=0x0100;}
reverseLED1(){PORT->P7^=0x0200;}

onLED0(){PORT->P7|=0x0100;}
onLED1(){PORT->P7|=0x0200;}	//Center Side
offLED0(){PORT->P7&=~0x0100;}
offLED1(){PORT->P7&=~0x0200;}	//Center Side

initializeTimer(){	/* OS Timer Interrupt */
	// 1st Stage
	ICDISR[4] =   0x00000000;	//1.(1)
	ICDICFR[8] =  0x5555F555;	//1.(2) 
	ICDIPR[33] =  0x00000000;	//1.(3) 00FF 0000
	ICDIPTR[33] = 0x00010000;	//1.(4) 0001 0000
	ICDISER[4] =  0x00000040;	//1.(5)
	// 2nd Stage
	ICC->PMR = 0x0F;
	ICC->BPR = 0x00;
	ICC->ICR = 0x03;
	// 3rd Stage
	*ICDDCR = 0x01;
	
//OS Timer Set
	OSTM0->TT = 0x01;	// Timer Start
	// OSTM0->EN = 	// 0x01 ;Counter Enable Register Read Only
	// OSTM0->TT = 0x01;	// Timer Stop
	OSTM0->CTL = 0x01;	// 0x01 Enable Interrupt ~0x02 interval timer
	OSTM0->CMP = 0xFFFFFF;	// 0xFFFF FFFF interval timer initial value 
	OSTM0->TS = 0x01;	// Timer Start
}

stopTimer(){
	ICDISER[4] =  0x00000000;	//1.(5)
	OSTM0->TT = 0x01;	// Timer Stop
}


/*	Interrupt Handler	*/
void enableInterrupt(void);
void disableInterrupt(void);

//static count =0;
void
__attribute__ ((interrupt("IRQ"))) 
interrupt(void) {// IRQ Handler for ARM CPU	
	int	prior, cause;
	
	prior = ICC->HPIR;
	cause = ICC->IAR;	//bit0-9 INT-ID 10-12 CPU ID


	if(cause == 134) {
		// OS Timer 0	Interrupt
	}
	if(cause == 235) { // SC3 RX Interrupt //	236 TX Interrupt
		putch(SC3, getch(SC3) );
	}
	if((cause == 233)|(cause ==234) ){ //BREAK || Frame Error
		SC3->SCFSR &= ~0x98;	//
	//	 0xF000	Parity Error Number	 0x0F00	Framing Error Number
	//	 0x80 ERror	 0x40 TEND	 0x20 TDFEmpty	 0x10 BRK
	//	 0x08 FERror	 0x04 PERror	 0x02 RDFull	 0x01 (Recieve)DReasy
		SC3->SCFCR = 00;		//
	}
	reverseLED0();

	ICC->EOIR = cause;

}

void
__attribute__ ((interrupt("FIQ")))
finterrupt(){ 
	putstr("FRQ\n");	//for(;;); 
}
void 
__attribute__ ((interrupt("SWI")))  
softwareInterrupt(){ 
	putstr("SWI\n");	//for(;;); 
}
void 
__attribute__ ((interrupt("ABORT")))  
prefetchAbort(){ 
	putstr("Prefetch\n");	//for(;;); 
}

void 
__attribute__ ((interrupt("ABORT")))  
dataAbort(){
	putstr("DataAbort\n");	//for(;;); 
}

void __attribute__ ((interrupt("UNDEF")))  
undefinedInstruction(){
	putstr("UNDEF\n");	//for(;;); 
}

